#define _CRT_SECURE_NO_DEPRECATE
#define _USE_MATH_DEFINES

#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>
#include "glwindow.h"
#include <GL/glu.h>
#include <algorithm>
#include <string>

#ifdef __linux__
#include <GL/glext.h>
#include <time.h>
#include <sys/time.h>
#else
#include "glext.h"
#endif

#include "shader.h"

/* environment setup */
static double alpha_water = 0.004;            // dB/km
static double alpha_sediment = 0.5000;        // dB/lambda
static double alpha_sponge = 0.01;
static double c0 = 1491.00;
static double c_sediment = 1700;
static double rho_water = 1000.0;
static double rho_sediment = 1500.0;
static double max_depth = 512.0;
static double max_range = 10240.0;
static double sponge_layer_start = 400.0;
static double zs = 25.0;     // center of initial source
static double zr = 75.0;     // receiver line
static unsigned num_freqs = 2;
static unsigned z_levels = 9;
static unsigned z_steps = (1 << z_levels);
static unsigned range_levels = 9;
static unsigned range_steps = (1 << range_levels);
static unsigned fft_cutoff = 128;   // at what size to begin using the twiddle support texture
static unsigned truncate_bits = 0;
static double delta_z = max_depth / z_steps;
static double delta_r = max_range / range_steps;
static unsigned window_width = 1024;
static unsigned window_height = 512;
static double delta_kz = 2.0 * M_PI / (2.0 * z_steps * delta_z);
static double z_nudge = 0.5;     // 0.0 = early discretization of z, 1.0 = late discretization
static double kz_nudge = 0.5;    // 0.0 = early discretization of k_z, 1.0 = late discretization
static bool no_display = false;
static bool enable_r300_hack = false;
static bool use_greene = false;
static GLuint precision_format = GL_RGBA_FLOAT32_ATI;

static std::string depth_filename = "data/depth.txt";
static std::string ssp_filename = "data/ssp.txt";
static std::string f0_filename = "data/f0.txt";
static std::string results_filename = "results.txt";
static std::string log_filename = "log.txt";

/* corresponding options */
const static struct option longopts[] = {
	{ "alpha-water", required_argument, NULL, 'a' },
	{ "alpha-sediment", required_argument, NULL, 'A' },
	{ "sound-speed-source", required_argument, NULL, 'c' },
	{ "sound-speed-sediment", required_argument, NULL, 'C' },
	{ "density-water", required_argument, NULL, 'e' },
	{ "density-sediment", required_argument, NULL, 'E' },
	{ "max-depth", required_argument, NULL, 'Z' },
	{ "max-range", required_argument, NULL, 'R' },
	{ "depth-res-order", required_argument, NULL, 'z' },
	{ "range-res-order", required_argument, NULL, 'r' },
	{ "sponge-layer-start", required_argument, NULL, 's' },
	{ "source-depth", required_argument, NULL, 'd' },
	{ "receiver-depth", required_argument, NULL, 'D' },
	{ "num-frequencies", required_argument, NULL, 'n' },
	{ "fft-cutoff", required_argument, NULL, 'f' },
	{ "window-width", required_argument, NULL, 'w' },
	{ "window-height", required_argument, NULL, 'h' },
	{ "z-nudge", required_argument, NULL, 'u' },
	{ "kz-nudge", required_argument, NULL, 'U' },
	{ "greene-source", no_argument, NULL, 'g' },
	{ "depth-file", required_argument, NULL, 'i' },
	{ "ssp-file", required_argument, NULL, 'j' },
	{ "f0-file", required_argument, NULL, 'k' },
	{ "result-file", required_argument, NULL, 'l' },
	{ "log-file", required_argument, NULL, 'm' },
	{ "force-16-bit", no_argument, NULL, '1' },
	{ "truncate-bits", no_argument, NULL, 't' },
	{ "r300-workaround", no_argument, NULL, '3' },
	{ "no-display", no_argument, NULL, 'q' },
	{ "help", no_argument, NULL, 'h' },
	{ NULL, 0, NULL, 0 }
};

float *start_field, *depth, *ssp, *f0, *support1, *support2, *delta_r_div_double_k0;
GLuint fb, fb_result;
GLuint depth_tex, ssp_tex, support1_tex, support2_tex, delta_r_div_double_k0_tex;
GLuint result_tex, initial_tex, rtt_tex1, rtt_tex2;
GLuint *twiddle_tex, *twiddle_inv_tex;
GLuint shuffle_tex, ravel_tex, unshuffle_tex, unravel_tex, exp_forward_tex, exp_inverse_tex;

Shader *fft_shader, *fft_constant_twiddle_shader, *display_shader, *refraction_shader, *diffraction_shader, *identity_shader, *truncate_shader, *identity_texcoord_shader;
Shader *shuffle_shader, *ravel_postmultiply_shader, *ravel_premultiply_shader;

struct fft_constant_data {
	double from0, to0;   // destination
	double from1, to1;   // source 1
	double from2, to2;   // source 2
	double twiddle_real, twiddle_imag;
};
std::vector<fft_constant_data> *twiddle_quads, *twiddle_inv_quads;

#ifdef __linux__
extern "C" void (*glXGetProcAddressARB(const GLubyte *procName))();
#endif

PFNGLGENFRAMEBUFFERSEXTPROC glGenFramebuffersEXT;
PFNGLGENRENDERBUFFERSEXTPROC glGenRenderbuffersEXT;
PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebufferEXT;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2DEXT;
PFNGLBINDRENDERBUFFEREXTPROC glBindRenderbufferEXT;
PFNGLRENDERBUFFERSTORAGEEXTPROC glRenderbufferStorageEXT;
PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatusEXT;
PFNGLCREATEPROGRAMOBJECTARBPROC glCreateProgramObjectARB;
PFNGLCREATESHADEROBJECTARBPROC glCreateShaderObjectARB;
PFNGLSHADERSOURCEARBPROC glShaderSourceARB;
PFNGLCOMPILESHADERARBPROC glCompileShaderARB;
PFNGLGETOBJECTPARAMETERIVARBPROC glGetObjectParameterivARB;
PFNGLGETINFOLOGARBPROC glGetInfoLogARB;
PFNGLATTACHOBJECTARBPROC glAttachObjectARB;
PFNGLDELETEOBJECTARBPROC glDeleteObjectARB;
PFNGLLINKPROGRAMARBPROC glLinkProgramARB;
PFNGLUSEPROGRAMOBJECTARBPROC glUseProgramObjectARB;
PFNGLGETUNIFORMLOCATIONARBPROC glGetUniformLocationARB;
PFNGLUNIFORM1IARBPROC glUniform1iARB;
PFNGLUNIFORM1FARBPROC glUniform1fARB;
PFNGLUNIFORM2FARBPROC glUniform2fARB;
PFNGLGETATTRIBLOCATIONARBPROC glGetAttribLocationARB;
PFNGLVERTEXATTRIB1FARBPROC glVertexAttrib1fARB;

#if WIN32 || !defined(GL_ARB_multitexture)
PFNGLMULTITEXCOORD1FARBPROC glMultiTexCoord1fARB;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB;
PFNGLMULTITEXCOORD3FARBPROC glMultiTexCoord3fARB;
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;
#endif

#define check_error() { GLenum ret = glGetError(); if (ret != 0) { printf("GL error at " __FILE__ ":%u: 0x%04x\n", __LINE__, ret); exit(1); } }

void simulate_truncation(GLuint tex, unsigned b);

#ifdef WIN32
int ffs(long i)
{
	return int(log2(float(i))) + 1;
}
#endif

void * get_proc_address(const char *proc)
{
#ifdef __linux__
	void *ret = (void *)glXGetProcAddressARB((const GLubyte *)proc);
#else
	void *ret = (void *)wglGetProcAddress(proc);
#endif
	if (ret == NULL) {
		fprintf(stderr, "Error: Couldn't find '%s'\n", proc);
	}

	return ret;
}

GLuint make_texture_1d(GLuint w, GLenum internal_format, GLenum format, GLfloat *data)
{
	GLuint tex;

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_1D, tex);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage1D(GL_TEXTURE_1D, 0, internal_format, w, 0, format, GL_FLOAT, data);
	check_error();

	return tex;
}

GLuint make_texture_2d(GLuint w, GLuint h, GLenum internal_format, GLenum format, GLfloat *data)
{
	GLuint tex;

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, internal_format, w, h, 0, format, GL_FLOAT, data);

	return tex;
}

/*
 * Writes 4/3 to the framebuffer in two different ways and measures
 * the precision; 4/3 is selected because its mantissa has every other
 * bit 1 and 0 in IEEE 754 notation, so it should catch truncation
 * errors quite quickly.
 */
void check_precision()
{
	float tmp[4];
	double test_double = 4.0/3.0;
	double error_texcoord, error_texture;

	glViewport(0, 0, z_steps, num_freqs/2);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rtt_tex2, 0);

	// check texcoord precision
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	{
		Shader::Binding sb(*identity_texcoord_shader);
		glBegin(GL_QUADS);

		glTexCoord1d(test_double);
		glVertex2f(0.0, 0.0);
		glVertex2f(1.0, 0.0);
		glVertex2f(1.0, 1.0);
		glVertex2f(0.0, 1.0);
		
		glEnd();
	}

	glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, tmp);
	error_texcoord = fabs(double(tmp[0]) - test_double);

	// check texture precision
	tmp[0] = test_double;
	GLuint test_tex = make_texture_2d(1, 1, precision_format, GL_RGBA, tmp);

	{
		Shader::Binding sb(*identity_shader);
		sb.bind_texture_2d("tex", test_tex);
		glBegin(GL_QUADS);

		glTexCoord1f(0.0);
		glVertex2f(0.0, 0.0);

		glTexCoord1f(1.0);
		glVertex2f(1.0, 0.0);;

		glTexCoord1f(1.0);
		glVertex2f(1.0, 1.0);

		glTexCoord1f(0.0);
		glVertex2f(0.0, 1.0);
		
		glEnd();
	}

	glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, tmp);
	error_texture = fabs(double(tmp[0]) - test_double);

	if (error_texture > 1e-07 && precision_format == GL_RGBA_FLOAT32_ATI) {
		printf("WARNING: Card does not have full 32-bit precision.\n");
		printf("         Expect suboptimal accuracy in results.\n");
	}
	if (error_texcoord/error_texture > 16.0) {
		printf("Note: Card has reduced texture coordinate precision.\n");
		printf("      Switching FFT strategy to compensate.\n");
		fft_cutoff = 8;
	}
}

void swap_buffers()
{
	if (truncate_bits > 0) {
		std::swap(rtt_tex1, rtt_tex2);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rtt_tex2, 0);
		simulate_truncation(rtt_tex1, truncate_bits);
	}

	std::swap(rtt_tex1, rtt_tex2);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rtt_tex2, 0);
}

void disp(GLuint tex, unsigned x)
{
	Shader::Binding sb(*display_shader);
	glViewport(0, 0, window_width, window_height);

	if (x == 0)
		glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, window_width, window_height, 0.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex);

	glBegin(GL_QUADS);

	glTexCoord2f(0.0, 1.0);
	glVertex2f(0.0, 0.0);
	
	glTexCoord2f(1.0, 1.0);
	glVertex2f(window_width, 0.0);
	
	glTexCoord2f(1.0, 0.0);
	glVertex2f(window_width, window_height);
	
	glTexCoord2f(0.0, 0.0);
	glVertex2f(0.0, window_height);
	
	glEnd();
}

/* 
 * Change this function and/or the appropriate shader, plus read_back() if you
 * want to store different data (say, data from more than the first frequency, or
 * a loss vs. range/frequency plot).
 */
void store(GLuint tex, unsigned x)
{
	Shader::Binding sb(*identity_shader);

	if (x == 0)
		glClear(GL_COLOR_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glViewport(0, 0, range_steps, z_steps);
	glOrtho(0.0, range_steps, z_steps, 0.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex);

	glBegin(GL_QUADS);

	glTexCoord2f(0.0, 0.0);
	glVertex2f(x, 0.0);
	
	glTexCoord2f(0.0, 1.0/double(num_freqs/2));
	glVertex2f(x + 1.0, 0.0);
	
	glTexCoord2f(1.0, 1.0/double(num_freqs/2));
	glVertex2f(x + 1.0, double(z_steps));
	
	glTexCoord2f(1.0, 0.0);
	glVertex2f(x, double(z_steps));
	
	glEnd();
}


void identity(GLuint tex)
{
	Shader::Binding sb(*identity_shader);
	sb.bind_texture_2d("tex", initial_tex);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	check_error();

	glBegin(GL_QUADS);

	glTexCoord2f(0.0, 0.0);
	glVertex2f(0.0, 0.0);
	
	glTexCoord2f(1.0, 0.0);
	glVertex2f(1.0, 0.0);
	
	glTexCoord2f(1.0, 1.0);
	glVertex2f(1.0, 1.0);
	
	glTexCoord2f(0.0, 1.0);
	glVertex2f(0.0, 1.0);
	
	glEnd();
}

void simulate_truncation(GLuint tex, unsigned b)
{
	Shader::Binding sb(*truncate_shader);
	sb.bind_texture_2d("tex", tex);
	sb.set_uniform("scalefac", float((1 << b) - 1));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	check_error();

	glBegin(GL_QUADS);

	glTexCoord2f(0.0, 0.0);
	glVertex2f(0.0, 0.0);
	
	glTexCoord2f(1.0, 0.0);
	glVertex2f(1.0, 0.0);
	
	glTexCoord2f(1.0, 1.0);
	glVertex2f(1.0, 1.0);
	
	glTexCoord2f(0.0, 1.0);
	glVertex2f(0.0, 1.0);
	
	glEnd();
}

void refraction(GLuint tex, unsigned frameno)
{
	Shader::Binding sb(*refraction_shader);
	sb.bind_texture_2d("tex", tex);
	sb.bind_texture_1d("support1", support1_tex);
	sb.bind_texture_1d("support2", support2_tex);
	sb.bind_texture_1d("depth", depth_tex);
	sb.bind_texture_1d("ssp", ssp_tex);

	sb.set_uniform("delta_r", delta_r);
	sb.set_uniform("r", frameno * delta_r);
	sb.set_uniform("r_coord", frameno / double(range_steps));
	sb.set_uniform("alpha_sponge", alpha_sponge);
	sb.set_uniform("rho_average", 0.5 * (rho_sediment + rho_water));
	sb.set_uniform("half_rho_difference", 0.5 * (rho_sediment - rho_water));
	sb.set_uniform("c_sediment", c_sediment);
	sb.set_uniform("c0_div_c_sediment_m1", c0 / c_sediment - 1.0);
	sb.set_uniform("c0", c0);
	sb.set_uniform("sponge_layer_start", sponge_layer_start);
	sb.set_uniform("sponge_layer_end", max_depth);
	sb.set_uniform("three_inv_sponge_len", 3.0 / (max_depth - sponge_layer_start));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glBegin(GL_QUADS);

	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0, 0.0);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 0.0);
	glMultiTexCoord2fARB(GL_TEXTURE2_ARB, z_nudge, 0.0);
	glVertex2f(0.0, 0.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0, 0.0);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 0.0);
	glMultiTexCoord2fARB(GL_TEXTURE2_ARB, max_depth + z_nudge, 1.0);
	glVertex2f(1.0, 0.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0, 1.0);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 1.0);
	glMultiTexCoord2fARB(GL_TEXTURE2_ARB, max_depth + z_nudge, 1.0);
	glVertex2f(1.0, 1.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0, 1.0);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 1.0);
	glMultiTexCoord2fARB(GL_TEXTURE2_ARB, z_nudge, 0.0);
	glVertex2f(0.0, 1.0);
	
	glEnd();
}

void shuffle(GLuint tex, GLuint shuffle_data_tex)
{
	Shader::Binding sb(*shuffle_shader);
	sb.bind_texture_2d("tex", tex);
	sb.bind_texture_1d("shuffle_data", shuffle_data_tex);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glBegin(GL_QUADS);

	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0, 0.0);
	glVertex2f(0.0, 0.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0, 0.0);
	glVertex2f(1.0, 0.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0, 1.0);
	glVertex2f(1.0, 1.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0, 1.0);
	glVertex2f(0.0, 1.0);
	
	glEnd();
}

void ravel(Shader &s, GLuint tex, GLuint rtex, GLuint etex)
{
	Shader::Binding sb(s);
	sb.bind_texture_2d("tex", tex);
	sb.bind_texture_1d("ravel_data_tex", rtex);
	sb.bind_texture_1d("exp_data_tex", etex);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glBegin(GL_QUADS);

	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0, 0.0);
	glVertex2f(0.0, 0.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0, 0.0);
	glVertex2f(1.0, 0.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0, 1.0);
	glVertex2f(1.0, 1.0);
	
	glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0, 1.0);
	glVertex2f(0.0, 1.0);
	
	glEnd();
}


void diffraction(GLuint tex)
{
	Shader::Binding sb(*diffraction_shader);
	sb.bind_texture_2d("tex", tex);
	sb.bind_texture_1d("delta_r_div_double_k0_tex", delta_r_div_double_k0_tex);
	sb.set_uniform("scalefac", 1.0 / double(z_steps));

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glBegin(GL_QUADS);

	glMultiTexCoord3fARB(GL_TEXTURE0_ARB, 0.0, 0.0, kz_nudge * delta_kz);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 0.0);
	glVertex2f(0.0, 0.0);
	
	glMultiTexCoord3fARB(GL_TEXTURE0_ARB, 1.0, 0.0, (z_steps + kz_nudge) * delta_kz);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 0.0);
	glVertex2f(1.0, 0.0);
	
	glMultiTexCoord3fARB(GL_TEXTURE0_ARB, 1.0, 1.0, (z_steps + kz_nudge) * delta_kz);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 1.0);
	glVertex2f(1.0, 1.0);
	
	glMultiTexCoord3fARB(GL_TEXTURE0_ARB, 0.0, 1.0, kz_nudge * delta_kz);
	glMultiTexCoord1fARB(GL_TEXTURE1_ARB, 1.0);
	glVertex2f(0.0, 1.0);
	
	glEnd();
}

GLuint make_twiddle_texture(unsigned fft_size, float mulfac, bool inverse)
{
	float tmp[z_steps * 4];
	float fft_stride = double(z_steps) / fft_size;

	for (unsigned element = 0; element < z_steps; ++element) {
		float k = floor(element / fft_stride);     // element # in this fft
		float offset = element - k * fft_stride;   // FFT number
		float twiddle_real, twiddle_imag;
		
		if (k < 0.5 * fft_size) {
			twiddle_real = cos(mulfac * k);
			twiddle_imag = sin(mulfac * k);
		} else {
			k -= 0.5 * fft_size;
			twiddle_real = -cos(mulfac * k);
			twiddle_imag = -sin(mulfac * k);
		}

		float base;
		base = k * fft_stride * 2.0 + offset;

		tmp[element * 4 + 0] = base / double(z_steps);
		tmp[element * 4 + 1] = (fft_stride + base) / double(z_steps);
		tmp[element * 4 + 2] = twiddle_real;
		tmp[element * 4 + 3] = twiddle_imag;
	}

	// convert this back to constant_sized quads
	std::vector<fft_constant_data> &q = (inverse) ? twiddle_inv_quads[ffs(fft_size) - 2] : twiddle_quads[ffs(fft_size) - 2];
	for (unsigned element = 0; element < z_steps; element += unsigned(fft_stride)) {
		fft_constant_data fd;

		fd.from0 = element;
		fd.to0 = (element + fft_stride);

		fd.from1 = 2 * element / double(z_steps);
		fd.to1 = (2 * element + fft_stride) / double(z_steps);

		fd.from2 = (2 * element + fft_stride) / double(z_steps);
		fd.to2 = (2 * element + 2 * fft_stride) / double(z_steps);

		fd.twiddle_real = tmp[element * 4 + 2];
		fd.twiddle_imag = tmp[element * 4 + 3];

		q.push_back(fd);
	}

	return make_texture_1d(z_steps, (fft_size <= 4) ? GL_RGBA_FLOAT16_ATI : precision_format, GL_RGBA, tmp);
}

// ravel/unravel textures are packed as [x1, x2, f1, f2], and the ravel
// shader computes f1 * x[x1] + i f2 x[x2]
GLuint make_ravel_texture(unsigned fft_size)
{
	float tmp[z_steps * 4];

	// y[0] = -2i x[N-1]
	tmp[0] = (fft_size - 1) / float(fft_size);
	tmp[1] = (fft_size - 1) / float(fft_size);
	tmp[2] = 0.0;
	tmp[3] = -2.0;

	// y[n] = 2x[n-1] - 2i x[N-n-1]      (n > 0)
	for (unsigned i = 1; i < unsigned(fft_size); ++i) {
		tmp[4*i + 0] = float(i - 1) / float(fft_size);
		tmp[4*i + 1] = (fft_size - i - 1) / float(fft_size);
		tmp[4*i + 2] = 2.0;
		tmp[4*i + 3] = -2.0;
	}

	return make_texture_1d(z_steps, precision_format, GL_RGBA, tmp);
}

GLuint make_unravel_texture(unsigned fft_size)
{
	float tmp[z_steps * 4];

	// x[n] = 1/4 ( y[n+1] + i y[N-n-1])      (n < N-1)
	for (unsigned i = 0; i < fft_size - 1; ++i) {
		tmp[4*i + 0] = (i + 1) / float(fft_size);
		tmp[4*i + 1] = (fft_size - i - 1) / float(fft_size);
		tmp[4*i + 2] = 0.25;
		tmp[4*i + 3] = 0.25;
	}

	// x[N-1] = 1/2 i y[0]
	tmp[4*(fft_size-1) + 0] = 0.0;
	tmp[4*(fft_size-1) + 1] = 0.0;
	tmp[4*(fft_size-1) + 2] = 0.0;
	tmp[4*(fft_size-1) + 3] = 0.5;

	return make_texture_1d(z_steps, precision_format, GL_RGBA, tmp);
}

// shuffle/unshuffle textures are packed as [x1, f1, 0, 0], and the shuffle
// shader computes f1 * x[x1]
GLuint make_shuffle_texture(unsigned fft_size)
{
	float tmp[z_steps * 4];

	for (unsigned i = 0; i < fft_size/2; ++i) {
		tmp[4*i + 0] = (i * 2) / float(fft_size);
		tmp[4*i + 1] = 1.0;
		tmp[4*i + 2] = 0.0;
		tmp[4*i + 3] = 0.0;
	}
	for (unsigned i = unsigned(fft_size)/2, j = 0; i < unsigned(fft_size); ++i, ++j) {
		tmp[4*i + 0] = (fft_size - j * 2 - 1) / float(fft_size);
		tmp[4*i + 1] = -1.0;
		tmp[4*i + 2] = 0.0;
		tmp[4*i + 3] = 0.0;
	}

	return make_texture_1d(z_steps, precision_format, GL_RGBA, tmp);
}

GLuint make_unshuffle_texture(unsigned fft_size)
{
	float tmp[z_steps * 4];

	for (unsigned i = 0; i < fft_size/2; ++i) {
		unsigned k = 2*i;

		tmp[4*k + 0] = i / float(fft_size);
		tmp[4*k + 1] = 1.0;
		tmp[4*k + 2] = 0.0;
		tmp[4*k + 3] = 0.0;
	}
	for (unsigned i = fft_size/2, j = 0; i < fft_size; ++i, ++j) {
		unsigned k = fft_size - 2*j - 1;

		tmp[4*k + 0] = i / float(fft_size);
		tmp[4*k + 1] = -1.0;
		tmp[4*k + 2] = 0.0;
		tmp[4*k + 3] = 0.0;
	}

	return make_texture_1d(z_steps, precision_format, GL_RGBA, tmp);
}

GLuint make_exp_texture(unsigned fft_size, bool inverse)
{
	float tmp[z_steps * 4];

	for (unsigned i = 0; i < fft_size; ++i) {
		double f = float(i)*M_PI/(2.0*fft_size);

		tmp[4*i + 0] = cos(f);
		tmp[4*i + 1] = (inverse ? -sin(f) : sin(f));
		tmp[4*i + 2] = 0.0;
		tmp[4*i + 3] = 0.0;
	}

	return make_texture_1d(z_steps, precision_format, GL_RGBA, tmp);
}

void precompute_twiddle()
{
	for (unsigned s = 2; s <= z_steps; s *= 2) {
		float mulfac = -2.0 * M_PI / double(s);
		twiddle_tex[ffs(s) - 2] = make_twiddle_texture(s, mulfac, false);
	}
	for (unsigned s = 2; s <= z_steps; s *= 2) {
		float mulfac = 2.0 * M_PI / double(s);
		twiddle_inv_tex[ffs(s) - 2] = make_twiddle_texture(s, mulfac, true);
	}
}

void fft(GLuint tex, GLuint size, bool inverse)
{	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, double(z_steps), 0.0, double(num_freqs/2), 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	Shader::Binding sb(*fft_shader);
	sb.bind_texture_2d("tex", tex);
	if (inverse)
		sb.bind_texture_1d("twiddle_tex", twiddle_inv_tex[ffs(size)-2]);
	else
		sb.bind_texture_1d("twiddle_tex", twiddle_tex[ffs(size)-2]);

	glBegin(GL_QUADS);

	glTexCoord2f(0.0, 0.0);
	glVertex2f(0.0, 0.0);
	
	glTexCoord2f(1.0, 0.0);
	glVertex2f(double(z_steps), 0.0);
	
	glTexCoord2f(1.0, 1.0);
	glVertex2f(double(z_steps), double(num_freqs/2));
	
	glTexCoord2f(0.0, 1.0);
	glVertex2f(0.0, double(num_freqs/2));

	glEnd();
}

void fft_constant_twiddle(GLuint tex, GLuint size, bool inverse)
{	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, double(z_steps), 0.0, double(num_freqs/2), 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	Shader::Binding sb(*fft_constant_twiddle_shader);
	sb.bind_texture_2d("tex", tex);

	std::vector<fft_constant_data> &q = (inverse) ? twiddle_inv_quads[ffs(size) - 2] : twiddle_quads[ffs(size) - 2];

	glBegin(GL_QUADS);

	for (std::vector<fft_constant_data>::const_iterator i = q.begin(); i != q.end(); ++i) {
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, i->from1, 0.0);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, i->from2, 0.0);
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, i->twiddle_real, i->twiddle_imag);
		glVertex2f(i->from0, 0.0);
		
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, i->to1, 0.0);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, i->to2, 0.0);
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, i->twiddle_real, i->twiddle_imag);
		glVertex2f(i->to0, 0.0);
		
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, i->to1, 1.0);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, i->to2, 1.0);
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, i->twiddle_real, i->twiddle_imag);
		glVertex2f(i->to0, double(num_freqs/2));
		
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, i->from1, 1.0);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, i->from2, 1.0);
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, i->twiddle_real, i->twiddle_imag);
		glVertex2f(i->from0, double(num_freqs/2));
	}

	glEnd();
}

GLint gen_rtt(GLenum format, GLuint width, GLuint height)
{
	GLuint rtt_tex;
	glGenTextures(1, &rtt_tex);
	
	glBindTexture(GL_TEXTURE_2D, rtt_tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rtt_tex, 0);

	check_error();
	return rtt_tex;
}

void read_depth(const char *filename, float *dst)
{
	FILE *f = fopen(filename, "r");
	if (f == NULL) {
		perror(filename);
		exit(1);
	}

	unsigned w;
	if (fscanf(f, "%u", &w) != 1) {
		fprintf(stderr, "short read in depth data\n");
		exit(1);
	}

	memset(dst, 0, range_steps * 4 * sizeof(float));
	
	for (unsigned x = 0; x < w && x < range_steps; ++x) {
		double d;
		if (fscanf(f, "%lf", &d) != 1) {
			fprintf(stderr, "short read in depth data\n");
			exit(1);
		}

		dst[x * 4] = d;
	}

	fclose(f);
}

void read_f0(const char *filename, float *dst1, float *dst2)
{
	FILE *f = fopen(filename, "r");
	if (f == NULL) {
		perror(filename);
		exit(1);
	}

	unsigned w;
	if (fscanf(f, "%u", &w) != 1) {
		fprintf(stderr, "short read in f0 data\n");
		exit(1);
	}

	memset(dst1, 0, num_freqs * 2 * sizeof(float));
	memset(dst2, 0, num_freqs * 2 * sizeof(float));
	memset(delta_r_div_double_k0, 0, num_freqs * 2 * sizeof(float));
	
	for (unsigned x = 0; x < w; ++x) {
		double g;

		if (fscanf(f, "%lf", &g) != 1) {
			fprintf(stderr, "short read in f0 data\n");
			exit(1);
		}

		if (x >= num_freqs)
			break;

		double k0 = 2.0 * M_PI * g / c0;

		f0[x] = g;
		dst1[x * 2 + 0] = (20.0 * M_LOG10E * alpha_water * ( c0 / ( g * 1000.0 ) )) / 27.29;
		dst1[x * 2 + 1] = alpha_sediment / 27.29;
		dst2[x * 2 + 0] = 0.5 * k0;
		dst2[x * 2 + 1] = 0.5 / (k0*k0);

		delta_r_div_double_k0[x * 2 + 0] = delta_r / (2.0*k0);
	}

	fclose(f);

}

// will neccessarily be incomplete (doesn't count the refraction layer)
// but that's OK; the other values will be ignored by the shader
void read_ssp(const char *filename, float *dst)
{
	FILE *f = fopen(filename, "r");
	if (f == NULL) {
		perror(filename);
		exit(1);
	}

	unsigned h;
	if (fscanf(f, "%u", &h) != 1) {
		fprintf(stderr, "short read in SSP data\n");
		exit(1);
	}

	memset(dst, 0, z_steps * 4 * sizeof(float));
	
	for (unsigned z = 0; z < h && z < z_steps; ++z) {
		double c;
		if (fscanf(f, "%lf", &c) != 1) {
			fprintf(stderr, "short read in SSP data\n");
			exit(1);
		}

		dst[z * 4] = c;
	}

	fclose(f);
}


void load_opengl_extensions()
{
	// FBO extension
	glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC)get_proc_address("glGenFramebuffersEXT");
	glGenRenderbuffersEXT = (PFNGLGENRENDERBUFFERSEXTPROC)get_proc_address("glGenRenderbuffersEXT");
	glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC)get_proc_address("glBindFramebufferEXT");
	glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)get_proc_address("glFramebufferTexture2DEXT");
	glBindRenderbufferEXT = (PFNGLBINDRENDERBUFFEREXTPROC)get_proc_address("glBindRenderbufferEXT");
	glRenderbufferStorageEXT = (PFNGLRENDERBUFFERSTORAGEEXTPROC)get_proc_address("glRenderbufferStorageEXT");
	glFramebufferRenderbufferEXT = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)get_proc_address("glFramebufferRenderbufferEXT");
	glCheckFramebufferStatusEXT = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC)get_proc_address("glCheckFramebufferStatusEXT");

	// shader extension
	glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)get_proc_address("glCreateProgramObjectARB");
	glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)get_proc_address("glCreateShaderObjectARB");
	glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)get_proc_address("glShaderSourceARB");
	glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)get_proc_address("glCompileShaderARB");
	glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)get_proc_address("glGetObjectParameterivARB");
	glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)get_proc_address("glGetInfoLogARB");
	glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)get_proc_address("glAttachObjectARB");
	glDeleteObjectARB = (PFNGLDELETEOBJECTARBPROC)get_proc_address("glDeleteObjectARB");
	glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)get_proc_address("glLinkProgramARB");
	glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)get_proc_address("glUseProgramObjectARB");
	glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)get_proc_address("glGetUniformLocationARB");
	glUniform1iARB = (PFNGLUNIFORM1IARBPROC)get_proc_address("glUniform1iARB");
	glUniform1fARB = (PFNGLUNIFORM1FARBPROC)get_proc_address("glUniform1fARB");
	glUniform2fARB = (PFNGLUNIFORM2FARBPROC)get_proc_address("glUniform2fARB");
	glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONARBPROC)get_proc_address("glGetAttribLocationARB");
	glVertexAttrib1fARB = (PFNGLVERTEXATTRIB1FARBPROC)get_proc_address("glVertexAttrib1fARB");
	
#if WIN32 || !defined(GL_ARB_multitexture)
	glMultiTexCoord1fARB = (PFNGLMULTITEXCOORD1FARBPROC)get_proc_address("glMultiTexCoord1fARB");
	glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC)get_proc_address("glMultiTexCoord2fARB");
	glMultiTexCoord3fARB = (PFNGLMULTITEXCOORD3FARBPROC)get_proc_address("glMultiTexCoord3fARB");
	glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)get_proc_address("glActiveTextureARB");
#endif

	check_error();
}

void generate_gaussian(float *start_field, float *freqs)
{
	for (unsigned f = 0; f < num_freqs; f += 2) {
		double k0_1 = 2.0 * M_PI * freqs[f] / c0;
		double k0_2 = 2.0 * M_PI * freqs[f+1] / c0;

		for (unsigned z = 0; z < z_steps; ++z) {
			float zf = float(z) * max_depth / double(z_steps);

			start_field[(z + (f/2) * z_steps) * 4 + 0] = sqrt(k0_1) * exp(-(k0_1 * k0_1 * .5) * (zf - zs) * (zf - zs));
			start_field[(z + (f/2) * z_steps) * 4 + 1] = 0.0;
			start_field[(z + (f/2) * z_steps) * 4 + 2] = sqrt(k0_2) * exp(-(k0_2 * k0_2 * .5) * (zf - zs) * (zf - zs));
			start_field[(z + (f/2) * z_steps) * 4 + 3] = 0.0;
		}
	}
}

void generate_greene(float *start_field, float *freqs)
{
	for (unsigned f = 0; f < num_freqs; f += 2) {
		double k0_1 = 2.0 * M_PI * freqs[f] / c0;
		double k0_2 = 2.0 * M_PI * freqs[f+1] / c0;

		for (unsigned z = 0; z < z_steps; ++z) {
			float zf = float(z) * max_depth / double(z_steps);

			start_field[(z + (f/2) * z_steps) * 4 + 0] = sqrt(k0_1) * (1.4467 - 0.4201*k0_1*k0_1 * (zf - zs) * (zf - zs))*exp(-(k0_1*(zf-zs)*(zf-zs))/3.0512);
			start_field[(z + (f/2) * z_steps) * 4 + 1] = 0.0;
			start_field[(z + (f/2) * z_steps) * 4 + 2] = sqrt(k0_2) * (1.4467 - 0.4201*k0_2*k0_2 * (zf - zs) * (zf - zs))*exp(-(k0_2*(zf-zs)*(zf-zs))/3.0512);
			start_field[(z + (f/2) * z_steps) * 4 + 3] = 0.0;
		}
	}
}

double abs_hankel0(double x)
{
	double real = j0(x), imag = y0(x);
	return sqrt(real * real + imag * imag);
}

void read_back()
{
	double k0 = 2.0 * M_PI * f0[0] / c0;
	unsigned z = (z_steps-1) - unsigned((zr - z_nudge) / delta_z + 0.5);
	float tmp[range_steps * 4];
		
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb_result);

	glReadPixels(0, z, range_steps, 1, GL_RGBA, GL_FLOAT, tmp);
	check_error();

	// All dB results are expected relative to r = 1m, but our results
	// are not correct that close to the origin. Let us just use
	// the Hankel function at 1m as base.
	float p1v = abs_hankel0(k0);

	FILE *f = fopen(results_filename.c_str(), "w");
	for (unsigned i = 0; i < range_steps; ++i) {
		float r = i * delta_r;
		float v = sqrt(tmp[i * 4] * tmp[i * 4] + tmp[i * 4 + 1] * tmp[i * 4 + 1]) * abs_hankel0(k0 * r);

		fprintf(f, "%f %f\n", r, -20.0*log10(v / p1v));
	}
	fclose(f);
}

/*
 * The (tab-separated) columns in the log file are:
 *
 * - Number of z-levels
 * - Number of range levels
 * - Number of frequencies
 * - FFT cutoff point
 * - Benchmark mode (no display) (0/1)
 * - Forced 16-bit precision (0/1)
 * - Truncate mode active (0/1)
 * - R300 hack (0/1)
 * - Time spent (seconds)
 * - Operating system (Win32/UNIX)
 * - OpenGL vendor
 * - OpenGL renderer
 * - OpenGL version
 */
void write_log(double elapsed)
{
	FILE *log = fopen(log_filename.c_str(), "a");
	
	fprintf(log, "%u\t%u\t%u\t%u\t%u\t%u\t%u\t%u\t%f\t",
		z_levels, range_levels, num_freqs, fft_cutoff,
		no_display, (precision_format == GL_RGBA_FLOAT16_ATI),
		(truncate_bits > 0), enable_r300_hack, elapsed);
#ifdef WIN32
	fprintf(log, "Win32\t");
#else
	fprintf(log, "UNIX\t");
#endif

	fprintf(log, "%s\t", glGetString(GL_VENDOR));
	fprintf(log, "%s\t", glGetString(GL_RENDERER));
	fprintf(log, "%s\n", glGetString(GL_VERSION));

	fclose(log);
}

void parse_options(int argc, char **argv)
{
	int option_index = 0;
	for ( ;; ) {

		int c = getopt_long(argc, argv, "a:A:c:C:e:E:Z:R:z:r:s:d:D:n:f:x:y:u:U:gi:j:k:l:m:q13t:h", longopts, &option_index);
		
		switch (c) {
		case 'a':
			alpha_water = atof(optarg);
			break;
		case 'A':
			alpha_sediment = atof(optarg);
			break;
		case 'c':
			c0 = atof(optarg);
			break;
		case 'C':
			c_sediment = atof(optarg);
			break;
		case 'e':
			rho_water = atof(optarg);
			break;
		case 'E':
			rho_sediment = atof(optarg);
			break;
		case 'Z':
			max_depth = atof(optarg);
			break;
		case 'R':
			max_range = atof(optarg);
			break;
		case 'z':
			z_levels = atoi(optarg);
			break;
		case 'r':
			range_levels = atoi(optarg);
			break;
		case 's':
			sponge_layer_start = atof(optarg);
			break;
		case 'd':
			zs = atof(optarg);
			break;
		case 'D':
			zr = atof(optarg);
			break;
		case 'n':
			num_freqs = atoi(optarg);
			if (num_freqs <= 0 || num_freqs % 2 != 0) {
				fprintf(stderr, "ERROR: Number of frequencies must be a positive, even number\n");
				exit(1);
			}
			break;
		case 'f':
			fft_cutoff = atoi(optarg);
			break;
		case 'x':
			window_width = atoi(optarg);
			break;
		case 'y':
			window_height = atoi(optarg);
			break;
		case 'u':
			z_nudge = atof(optarg);
			break;
		case 'U':
			kz_nudge = atof(optarg);
			break;
		case 'g':
			use_greene = true;
			break;
		case 'q':
			no_display = true;
			break;
		case 'i':
			depth_filename = optarg;
			break;
		case 'j':
			ssp_filename = optarg;
			break;
		case 'k':
			f0_filename = optarg;
			break;
		case 'l':
			results_filename = optarg;
			break;
		case 'm':
			log_filename = optarg;
			break;
		case '1':
			precision_format = GL_RGBA_FLOAT16_ATI;
			break;
		case '3':
			enable_r300_hack = true;
			break;
		case 't':
			truncate_bits = atoi(optarg);
			break;
		case -1:		// end of argument list
			return;
		default:                // --help, or invalid option
			fprintf(stderr, "Usage: gpuwave [OPTION]...\n\n");
			fprintf(stderr, "  -a, --alpha-water=VAL           set dampening in water (dB/km), default %f\n", alpha_water);
			fprintf(stderr, "  -A, --alpha-sediment=VAL        set dampening in sediment (dB/lambda), default %f\n", alpha_sediment);
			fprintf(stderr, "  -c, --sound-speed-source=VAL    set sound speed at the source (m/s), default %f\n", c0);
			fprintf(stderr, "  -C, --sound-speed-sediment=VAL  set sound speed in sediment (m/s), default %f\n", c_sediment);
			fprintf(stderr, "  -e, --density-water=VAL         set density of water (kg/m^3), default %f\n", rho_water);
			fprintf(stderr, "  -E, --density-sediment=VAL      set density of sediment (kg/m^3), default %f\n", rho_sediment);
			fprintf(stderr, "  -Z, --max-depth=VAL             set bottom of simulation area (m), default %f\n", max_depth);
			fprintf(stderr, "  -R, --max-range=VAL             set range of simulation area (m), default %f\n", max_range);
			fprintf(stderr, "  -z, --depth-res-order=VAL       depth resolution order (depth = 2^z), default %u\n", z_levels);
			fprintf(stderr, "  -r, --range-res-order=VAL       range resolution order (range = 2^r), default %u\n", range_levels);
			fprintf(stderr, "  -s, --sponge-layer-start=VAL    start depth of sponge layer (m), default %f\n", sponge_layer_start);
			fprintf(stderr, "  -d, --source-depth=VAL          depth of source (m), default %f\n", zs);
			fprintf(stderr, "  -D, --receiver-depth=VAL        depth of receiver (m), default %f\n", zr);
			fprintf(stderr, "  -n, --num-frequencies=VAL       number of frequencies to calculate for, default %u\n", num_freqs);
			fprintf(stderr, "  -f, --fft-cutoff=VAL            FFT size strategy crossover point, default %u\n", fft_cutoff);
			fprintf(stderr, "  -x, --window-width=VAL          width of display window, default %u\n", window_width);
			fprintf(stderr, "  -y, --window-height=VAL         height of display window, default %u\n", window_height);
			fprintf(stderr, "  -u, --z-nudge=VAL               z discretization adjustment [0..1], default %f\n", z_nudge);
			fprintf(stderr, "  -U, --kz-nudge=VAL              k_z discretization adjustment [0..1], default %f\n", kz_nudge);
			fprintf(stderr, "  -g, --greene-source             Use Greene's source instead of Gaussian source\n");
			fprintf(stderr, "\n");
			fprintf(stderr, "  -i, --depth-file=FILENAME       file containing depth data (default %s)\n", depth_filename.c_str());
			fprintf(stderr, "  -j, --ssp-file=FILENAME         file containing the sound speed profile (default %s)\n", ssp_filename.c_str());
			fprintf(stderr, "  -k, --f0-file=FILENAME          file containing source frequencies (default %s)\n", f0_filename.c_str());
			fprintf(stderr, "  -l, --results-file=FILENAME     file to write results to (default %s)\n", results_filename.c_str());
			fprintf(stderr, "  -m, --log-file=FILENAME         file to write performance log to (default %s)\n", log_filename.c_str());
			fprintf(stderr, "\n");
			fprintf(stderr, "  -q, --no-display                turn off display (for benchmarking)\n");
			fprintf(stderr, "  -1, --force-16-bit              force 16-bit float precision\n");
			fprintf(stderr, "  -3, --r300-workaround           hack refraction shader for R300 (benchmarking only!)\n");
			fprintf(stderr, "  -t, --truncate-bits=NUM         simulate N bits of mantissa truncation (default %u)\n", truncate_bits);
			fprintf(stderr, "  -h, --help                      display this help and exit\n");
			exit(1);
		}
	}
}

int main(int argc, char **argv)
{
	parse_options(argc, argv);
	GLWindow *win = new GLWindow("Underwater acoustics simulation", window_width, window_height, 32, false);
	
	load_opengl_extensions();

	// these may have changed due to options
	z_steps = 1 << z_levels;
	range_steps = 1 << range_levels;
	delta_z = max_depth / z_steps;
	delta_r = max_range / range_steps;
	delta_kz = 2.0 * M_PI / (2.0 * z_steps * delta_z);

	// allocate memory
	start_field = new float[z_steps * num_freqs/2 * 4];
	depth = new float[range_steps * 4];
	ssp = new float[z_steps * 4];
	f0 = new float[num_freqs];
	support1 = new float[num_freqs/2 * 4];
	support2 = new float[num_freqs/2 * 4];
	delta_r_div_double_k0 = new float[num_freqs/2 * 4];
	twiddle_tex = new GLuint[z_levels];
	twiddle_inv_tex = new GLuint[z_levels];
	twiddle_quads = new std::vector<fft_constant_data> [z_levels];
	twiddle_inv_quads = new std::vector<fft_constant_data> [z_levels];

	// load shaders
	fft_shader = new Shader("fft-v.glsl", "fft-f.glsl");
	fft_constant_twiddle_shader = new Shader("fftc-v.glsl", "fftc-f.glsl");
	display_shader = new Shader("identity-v.glsl", "display-f.glsl");
	refraction_shader = new Shader("refraction-v.glsl", "refraction-f.glsl", enable_r300_hack);
	diffraction_shader = new Shader("diffraction-v.glsl", "diffraction-f.glsl");
	identity_shader = new Shader("identity-v.glsl", "identity-f.glsl");
	truncate_shader = new Shader("identity-v.glsl", "truncate-f.glsl");
	identity_texcoord_shader = new Shader("identity-v.glsl", "identity-texcoord-f.glsl");
	shuffle_shader = new Shader("shuffle-v.glsl", "shuffle-f.glsl");
	ravel_postmultiply_shader = new Shader("ravel-v.glsl", "ravel-postmultiply-f.glsl");
	ravel_premultiply_shader = new Shader("ravel-v.glsl", "ravel-premultiply-f.glsl");

	// generate our framebuffers
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	rtt_tex1 = gen_rtt(precision_format, z_steps, num_freqs/2);
	rtt_tex2 = gen_rtt(precision_format, z_steps, num_freqs/2);
	
	glGenFramebuffersEXT(1, &fb_result);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb_result);
	result_tex = gen_rtt(precision_format, range_steps, z_steps);

	check_error();
	
	{                                                           
		GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT); 
		switch(status) {                                          
		case GL_FRAMEBUFFER_COMPLETE_EXT:                       
			break;                                                
		case GL_FRAMEBUFFER_UNSUPPORTED_EXT:                    
			printf("configuration unsupported\n");
			exit(1);
		default:
			printf("unknown status %x\n", status);
			exit(1);
		}
	}

	// depth (in the ocean sense, not in the OpenGL sense) texture
	read_depth(depth_filename.c_str(), depth);
	depth_tex = make_texture_1d(range_steps, precision_format, GL_RGBA, depth);
	
	// sound speed profile texture
	read_ssp(ssp_filename.c_str(), ssp);
	ssp_tex = make_texture_1d(z_steps, precision_format, GL_RGBA, ssp);

	// f0 data
	read_f0(f0_filename.c_str(), support1, support2);
	support1_tex = make_texture_1d(num_freqs/2, precision_format, GL_RGBA, support1);
	support2_tex = make_texture_1d(num_freqs/2, precision_format, GL_RGBA, support2);
	delta_r_div_double_k0_tex = make_texture_1d(num_freqs/2, precision_format, GL_RGBA, delta_r_div_double_k0);
	
	// start field texture,
	if (use_greene)
		generate_greene(start_field, f0);
	else
		generate_gaussian(start_field, f0);
	initial_tex = make_texture_2d(z_steps, num_freqs/2, precision_format, GL_RGBA, start_field);
	
	precompute_twiddle();
	ravel_tex = make_ravel_texture(z_steps);
	unravel_tex = make_unravel_texture(z_steps);
	shuffle_tex = make_shuffle_texture(z_steps);
	unshuffle_tex = make_unshuffle_texture(z_steps);
	exp_forward_tex = make_exp_texture(z_steps, false);
	exp_inverse_tex = make_exp_texture(z_steps, true);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

	check_precision();

	// input the starting field
	glViewport(0, 0, z_steps, num_freqs/2);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rtt_tex2, 0);
	
	identity(initial_tex);
	swap_buffers();
	check_error();

#if WIN32
	DWORD start, end;
	start = timeGetTime();
#else
	struct timeval start, end;
	gettimeofday(&start, NULL);
#endif

	while (!win->is_done()) {
		static unsigned frameno = 0;
	
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, rtt_tex2, 0);
		glViewport(0, 0, z_steps, num_freqs/2);
		
		// convert DST to IFFT (preprocessing)
		shuffle(rtt_tex1, shuffle_tex);
		swap_buffers();

		// inverse FFT
		for (unsigned s = 2; s <= z_steps; s *= 2) {
			if (s < fft_cutoff) 
				fft_constant_twiddle(rtt_tex1, s, true);
			else 
				fft(rtt_tex1, s, true);
			swap_buffers();
		}

		// and postprocessing (forward exp, then unravel) to finish the DST
		ravel(*ravel_premultiply_shader, rtt_tex1, unravel_tex, exp_forward_tex);
		swap_buffers();

		// multiply by diffraction term
		diffraction(rtt_tex1);
		swap_buffers();

		// convert IDST to FFT (preprocessing: ravel, then inverse exp)
		ravel(*ravel_postmultiply_shader, rtt_tex1, ravel_tex, exp_inverse_tex);
		swap_buffers();

		// forward FFT
		for (unsigned s = 2; s <= z_steps; s *= 2) {
			if (s < fft_cutoff)
				fft_constant_twiddle(rtt_tex1, s, false);
			else 
				fft(rtt_tex1, s, false);
			swap_buffers();
		}

		// and finally unshuffle to finish the IDST
		shuffle(rtt_tex1, unshuffle_tex);
		swap_buffers();

		refraction(rtt_tex1, frameno);
		swap_buffers();

		if (!no_display) {
			// store
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb_result);
			store(rtt_tex1, frameno);
			check_error();

			// and display on screen
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
			disp(result_tex, frameno);
			
			win->flip();
		}
		
		++frameno;
		if (frameno == range_steps)
			break;
	}
	
	double elapsed;
#if WIN32
	end = timeGetTime();
	elapsed = double(end - start) * 1e-3;
#else
	gettimeofday(&end, NULL);
	elapsed = double(end.tv_sec - start.tv_sec) +
		double(end.tv_usec - start.tv_usec) * 1e-6;
#endif

	glFinish();
	printf("Simulation took %.3f seconds.\n", elapsed);

	write_log(elapsed);
	read_back();	

	delete win;
}
