#include "shader.h"

extern PFNGLCREATEPROGRAMOBJECTARBPROC glCreateProgramObjectARB;
extern PFNGLCREATESHADEROBJECTARBPROC glCreateShaderObjectARB;
extern PFNGLSHADERSOURCEARBPROC glShaderSourceARB;
extern PFNGLCOMPILESHADERARBPROC glCompileShaderARB;
extern PFNGLGETOBJECTPARAMETERIVARBPROC glGetObjectParameterivARB;
extern PFNGLGETINFOLOGARBPROC glGetInfoLogARB;
extern PFNGLATTACHOBJECTARBPROC glAttachObjectARB;
extern PFNGLDELETEOBJECTARBPROC glDeleteObjectARB;
extern PFNGLLINKPROGRAMARBPROC glLinkProgramARB;
extern PFNGLUSEPROGRAMOBJECTARBPROC glUseProgramObjectARB;
extern PFNGLGETUNIFORMLOCATIONARBPROC glGetUniformLocationARB;
extern PFNGLUNIFORM1IARBPROC glUniform1iARB;
extern PFNGLUNIFORM1FARBPROC glUniform1fARB;
extern PFNGLUNIFORM2FARBPROC glUniform2fARB;
extern PFNGLGETATTRIBLOCATIONARBPROC glGetAttribLocationARB;
extern PFNGLVERTEXATTRIB1FARBPROC glVertexAttrib1fARB;

#if WIN32 || !defined(GL_ARB_multitexture)
extern PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;
#endif

Shader::Shader(const std::string &vs_filename, const std::string &fs_filename, bool enable_r300_hack)
	: enable_r300_hack(enable_r300_hack)
{
	prog = glCreateProgramObjectARB();
	GLuint vs = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	GLuint fs = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

	compile_shader(vs, "shaders/" + vs_filename);
	compile_shader(fs, "shaders/" + fs_filename);
	glAttachObjectARB(prog, vs);
	glAttachObjectARB(prog, fs);
	glDeleteObjectARB(vs);
	glDeleteObjectARB(fs);

	glLinkProgramARB(prog);
	check_linked();
}

Shader::~Shader()
{
//	glDeleteProgramObjectARB(prog);
}

void Shader::compile_shader(GLuint obj, const std::string &filename)
{
	FILE *file = fopen(filename.c_str(), "r");
	if (file == NULL) {
		perror(filename.c_str());
		exit(1);
	}

	GLcharARB *buf = (GLcharARB *)malloc(16384);
	GLint len = fread(buf, 1, 16384, file);

	char define_str[256];
	sprintf(define_str, "#define R300_BENCHMARK_WORKAROUND %u\n", enable_r300_hack);

	const GLcharARB* ptrs[2] = { define_str, buf };
	const GLint lens[2] = { strlen(define_str), len };
	glShaderSourceARB(obj, 2, ptrs, lens);

	glCompileShaderARB(obj);
	free(buf);

	GLint compiled = GL_FALSE;
	GLint maxLength, length;
	glGetObjectParameterivARB(obj, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
	glGetObjectParameterivARB(obj, GL_OBJECT_INFO_LOG_LENGTH_ARB, &maxLength);
	GLcharARB *pInfoLog = (GLcharARB *) malloc(maxLength * sizeof(GLcharARB));
	glGetInfoLogARB(obj, maxLength, &length, pInfoLog);

	if (!compiled) {
		printf("Compile of '%s' failed: %s\n", filename.c_str(), pInfoLog);
		exit(1);
	}
	free(pInfoLog);
}

void Shader::check_linked()
{
	GLint linked = GL_FALSE;
	GLint maxLength, length;
	glGetObjectParameterivARB(prog, GL_OBJECT_LINK_STATUS_ARB, &linked);
	glGetObjectParameterivARB(prog, GL_OBJECT_INFO_LOG_LENGTH_ARB, &maxLength);
	GLcharARB *pInfoLog = (GLcharARB *) malloc(maxLength * sizeof(GLcharARB));
	glGetInfoLogARB(prog, maxLength, &length, pInfoLog);

	if (!linked) {
		printf("%s\n", pInfoLog);
		exit(1);
	}
	free(pInfoLog);
}

Shader::Binding::Binding(Shader &s)
 	: s(s), current_texunit(0)
{
	glUseProgramObjectARB(s.prog);
}

Shader::Binding::~Binding()
{
	for (std::vector<std::pair<GLenum, GLenum> >::const_iterator i = active_bindings.begin(); i != active_bindings.end(); ++i) {
		glActiveTextureARB(i->second);
		glDisable(i->first);
	}
	glActiveTextureARB(GL_TEXTURE0_ARB);
}

void Shader::Binding::bind_texture_1d(const std::string &name, GLuint tex)
{
	glActiveTextureARB(GL_TEXTURE0_ARB + current_texunit);
	glEnable(GL_TEXTURE_1D);
	glBindTexture(GL_TEXTURE_1D, tex);
	glUniform1iARB(glGetUniformLocationARB(s.prog, name.c_str()), current_texunit);

	active_bindings.push_back(std::make_pair(GL_TEXTURE_1D, GL_TEXTURE0_ARB + current_texunit));
	++current_texunit;
}

void Shader::Binding::bind_texture_2d(const std::string &name, GLuint tex)
{
	glActiveTextureARB(GL_TEXTURE0_ARB + current_texunit);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex);
	glUniform1iARB(glGetUniformLocationARB(s.prog, name.c_str()), current_texunit);

	active_bindings.push_back(std::make_pair(GL_TEXTURE_2D, GL_TEXTURE0_ARB + current_texunit));
	++current_texunit;
}

void Shader::Binding::set_uniform(const std::string &name, GLfloat x)
{
	glUniform1fARB(glGetUniformLocationARB(s.prog, name.c_str()), x);
}

