uniform sampler2D tex;
varying vec2 tex_coord;

uniform sampler1D depth;
uniform sampler1D ssp;              // sound speed profile
uniform sampler1D support1;         // [alpha_water,alpha_sediment] for both f0
uniform sampler1D support2;         // [k0,half_inv_k0_squared]     for both f0
uniform float alpha_sponge;
uniform float c_sediment;
uniform float c0_div_c_sediment_m1; // (c0 / c_sediment) - 1
uniform float rho_average;          // 1/2 (rho_sediment + rho_water)
uniform float half_rho_difference;  // 1/2 (rho_sediment - rho_water)
uniform float c0;
uniform float r;
uniform float r_coord;
uniform float sponge_layer_start;
uniform float sponge_layer_end;
uniform float delta_r;
uniform float three_inv_sponge_len;   // 3.0 / (sponge_layer_end - sponge_layer_start)
varying float z;
varying float z_coord;
varying float f0_coord;

vec2 complex_multiply(in vec2 a, in vec2 b)
{
	return a.x * b.xy + a.y * vec2(-b.y, b.x);
}

// computes exp(i(x + yi))
vec2 exp_p(vec2 z)
{
	float e = exp(-z.y);
	return e * vec2(cos(z.x), sin(z.x));
}

void main(void)
{
	vec4 color = texture2D(tex, tex_coord);
	vec4 n2;
	vec4 supp1 = texture1D(support1, f0_coord);
	vec4 supp2 = texture1D(support2, f0_coord);

	if (z >= sponge_layer_start) {
		float temp = (z - sponge_layer_start) * three_inv_sponge_len;

		n2.xz = vec2(c0_div_c_sediment_m1);
		n2.yw = vec2(alpha_sponge * exp(temp*temp));
	} else {
		vec2 alpha;
		float c = texture1D(ssp, z_coord).x;
		float D = texture1D(depth, r_coord).x;

		//
		// The compiler will normally change this kind of construct
		// from branches to conditional writes -- no need to get ugly
		// with mix() and step().
		//
		if (z >= D) {
			c = c_sediment;
			alpha = supp1.yw;
		} else {
			alpha = supp1.xz;
		}

		// density
		vec2 tanh_arg_double = 2.0 * supp2.xz * (z - D);

		// this is needed to keep the exp() from growing without bounds --
		// needed on some cards
		tanh_arg_double = clamp(tanh_arg_double, -20.0, 20.0);

		vec2 th;
		{
			float e2x = exp(tanh_arg_double.x);        // exp(2x)
			vec2 temp = vec2(e2x) + vec2(-1.0, 1.0);   // exp(2x)-1, exp(2x)+1
			th.x = temp.x / temp.y;                    // tanh(x)
		}
		{
#if R300_BENCHMARK_WORKAROUND
			// use this version for the R300 -- it gives wrong answers, but the
			// computational workload is comparable, and the correct version
			// segfaults the GLSL compiler
			float e2x = exp(tanh_arg_double.x);        // exp(2x)
#else
			float e2x = exp(tanh_arg_double.y);        // exp(2x)
#endif
			vec2 temp = vec2(e2x) + vec2(-1.0, 1.0);   // exp(2x)-1, exp(2x)+1
			th.y = temp.x / temp.y;                    // tanh(x)
		} 
		vec2 th_dx = 1.0 - th*th;             // 1 - tanh(x)^2
		vec2 th_dxdx = -2.0 * th * th_dx;     // -2 tanh(x) (1 - tanh(x)^2)

		// note that (tanh(A*x))' = A*tanh(A*x), thus the extra muls by supp2.xz
		vec2 rho = vec2(rho_average) + vec2(half_rho_difference) * th;
		vec2 rho_dx = supp2.xz * vec2(half_rho_difference) * th_dx;
		vec2 rho_dxdx = supp2.xz * supp2.xz * vec2(half_rho_difference) * th_dxdx;
		vec2 rho_inv = 1.0 / rho;

		// now finally calculate n^2
		float n = c0 / c;
		n = n * n;

		n2.xz = n + supp2.yw * rho_inv * (rho_dxdx - 1.5 * rho_inv * rho_dx * rho_dx) - 1.0;
		n2.yw = n * alpha;
	}
	
	vec4 A = n2 * delta_r * supp2.xxzz;
	gl_FragColor.rg = complex_multiply(exp_p(A.xy), color.rg);
	gl_FragColor.ba = complex_multiply(exp_p(A.zw), color.ba);
}

